#include <WiFi.h>
#include <WebServer.h>
#include "html.h"
#include <Firebase_ESP_Client.h>

// Provide the token generation process info.
#include "addons/TokenHelper.h"
// Provide the RTDB payload printing info and other helper functions.
#include "addons/RTDBHelper.h"

// Insert Firebase project API Key
#define API_KEY "AIzaSyA-Ab3684hy2PrG1M5gPLShQG1pR9Mc4nE"

// Insert RTDB URLefine the RTDB URL */
#define DATABASE_URL "iotdb-87405-default-rtdb.firebaseio.com" 

WebServer server(80);
// innitializing ports of soil moisture sensor
int _moisture,sensor_analog;
const int sensor_pin = A0;
const char* ssid = "SLT FIBRE HOME";
const char* password = "19620925";

FirebaseData fbdo;

FirebaseAuth auth;
FirebaseConfig config;
FirebaseData stream;
FirebaseData firebaseData;
FirebaseJson json;
bool isAuthenticated = false;
// Firebase database path
String databasePath = ""; 
// Firebase Unique Identifier
String fuid = "";

// Variables to save database paths
String listenerPath = "board1/outputs/digital/";
const int output1 = 2;

String led_state = "";                               
int led_gpio = 2;

// Timer variables (send new readings every three secnd)
unsigned long sendDataPrevMillis = 0;
unsigned long timerDelay = 3000;
int timestamp;
unsigned long getTime() {
  time_t now;
  struct tm timeinfo;
  if (!getLocalTime(&timeinfo)) {
    //Serial.println("Failed to obtain time");
    return(0);
  }
  time(&now);
  return now;
}

void MainPage() {
  String _html_page = html_page;              /*Read The HTML Page*/
  server.send(200, "text/html", _html_page);  /*Send the code to the web server*/
}

void SoilMoisture() {
  String MoistuValue = String(_moisture);    //Convert it into string
  server.send(200, "text/plane", MoistuValue);  //Send updated temperature value to the web server
}
// connecting to the wifi
void initWiFi() {
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  Serial.print("Connecting to WiFi ..");
   Serial.println(ssid);

  while (WiFi.waitForConnectResult() != WL_CONNECTED) {
    Serial.print('.');
    delay(100);
  }
  Serial.print("Connected to ");
  Serial.println(ssid);
  Serial.print("Your Local IP address is: ");
  Serial.println(WiFi.localIP()); 
}

// Callback function that runs on database changes
void streamCallback(FirebaseStream data){
  Serial.printf("stream path, %s\nevent path, %s\ndata type, %s\nevent type, %s\n\n",
                data.streamPath().c_str(),
                data.dataPath().c_str(),
                data.dataType().c_str(),
                data.eventType().c_str());
  printResult(data); //see addons/RTDBHelper.h
  Serial.println();

  // Get the path that triggered the function
  String streamPath = String(data.dataPath());

  // if the data returned is an integer, there was a change on the GPIO state on the following path /{gpio_number}
  if (data.dataTypeEnum() == fb_esp_rtdb_data_type_integer){
    String gpio = streamPath.substring(1);
    int state = data.intData();
    Serial.print("GPIO: ");
    Serial.println(gpio);
    Serial.print("STATE: ");
    Serial.println(state);
    digitalWrite(gpio.toInt(), state);
  }

  /* When it first runs, it is triggered on the root (/) path and returns a JSON with all keys
  and values of that path. So, we can get all values from the database and updated the GPIO states*/
  if (data.dataTypeEnum() == fb_esp_rtdb_data_type_json){
    FirebaseJson json = data.to<FirebaseJson>();

    // To iterate all values in Json object
    size_t count = json.iteratorBegin();
    Serial.println("\n---------");
    for (size_t i = 0; i < count; i++){
        FirebaseJson::IteratorValue value = json.valueAt(i);
        int gpio = value.key.toInt();
        int state = value.value.toInt();
        Serial.print("STATE: ");
        Serial.println(state);
        Serial.print("GPIO:");
        Serial.println(gpio);
        digitalWrite(gpio, state);
        Serial.printf("Name: %s, Value: %s, Type: %s\n", value.key.c_str(), value.value.c_str(), value.type == FirebaseJson::JSON_OBJECT ? "object" : "array");
    }
    Serial.println();
    json.iteratorEnd(); // required for free the used memory in iteration (node data collection)
  }
  
  //This is the size of stream payload received (current and max value)
  //Max payload size is the payload size under the stream path since the stream connected
  //and read once and will not update until stream reconnection takes place.
  //This max value will be zero as no payload received in case of ESP8266 which
  //BearSSL reserved Rx buffer size is less than the actual stream payload.
  Serial.printf("Received stream payload size: %d (Max. %d)\n\n", data.payloadLength(), data.maxPayloadLength());
}

void streamTimeoutCallback(bool timeout){
  if (timeout)
    Serial.println("stream timeout, resuming...\n");
  if (!stream.httpConnected())
    Serial.printf("error code: %d, reason: %s\n\n", stream.httpCode(), stream.errorReason().c_str());
}

void setup() {
  Serial.begin(115200);
  initWiFi();
  pinMode(output1, OUTPUT); 

  config.api_key = API_KEY;
  config.database_url = DATABASE_URL;

  Firebase.reconnectWiFi(true);

  if (Firebase.signUp(&config, &auth, "", "")){
    Serial.println("Success");
    isAuthenticated = true;
    // Set the database path where updates will be loaded for this device
    databasePath = "/main";
    fuid = auth.token.uid.c_str();
  }else{
    Serial.printf("Failed, %s\n", config.signer.signupError.message.c_str());
    isAuthenticated = false;
  }
  Firebase.begin(&config,&auth);
  // Streaming (whenever data changes on a path)
  // Begin stream on a database path --> board1/outputs/digital
  if (!Firebase.RTDB.beginStream(&stream, listenerPath.c_str()))
    Serial.printf("stream begin error, %s\n\n", stream.errorReason().c_str());  
  // Firebase.reconnectWiFi(true);

   // Assign a calback function to run when it detects changes on the database
  Firebase.RTDB.setStreamCallback(&stream, streamCallback, streamTimeoutCallback); 

  // Firebase.begin(FIREBASE_HOST,FIREBASE_Authorization_key);
  server.on("/", MainPage);           /*Display the Web/HTML Page*/
  server.on("/readMoisture", SoilMoisture);       /*Display the updated Temperature and Humidity value*/
  server.begin();                    /*Start Server*/
  delay(1000);      
}
// To read data
void loop() {
  if (Firebase.ready() && (millis() - sendDataPrevMillis > timerDelay || sendDataPrevMillis == 0)){
    sendDataPrevMillis = millis();
        Firebase.refreshToken(&config);
    Serial.println("Refresh token");
  sensor_analog = analogRead(sensor_pin);
  _moisture = ( 100 - ( (sensor_analog/4095.00) * 100 ) );
  timestamp = getTime();
  Serial.print ("time: ");
  Serial.println (timestamp);
  server.handleClient(); 
  Serial.print("Moisture = ");
  Serial.print(_moisture);           /* Print Temperature on the serial window */
  Serial.println("%");
  json.set(timestamp, String(_moisture));
  Serial.printf("Set json... %s\n", Firebase.RTDB.setJSON(&fbdo, "/test", &json) ? "ok" : fbdo.errorReason().c_str());
  }
  // delay(1000); 
}
